// ntfsattributes.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include "ntfsattributes.h"
#include "ntfsutil.h"

#include "ntfsfile.h"
#include "ntfsattributelistentry.h"

NTFSAttributes::NTFSAttributes()
{
	m_list.setAutoDelete(TRUE);
}

NTFSAttributes::~NTFSAttributes()
{
}

void NTFSAttributes::zero()
{
	m_list.clear();
}

bool NTFSAttributes::read(NTFSStream * s)
{
	NTFSAttribute * a;
	NTFSStream ss;

	// Read attributes from stream, up to maxLen.
	while (s->getOffset() < s->getLength())
	{
		// Create a new attribute
		a = new NTFSAttribute();

		// Set ss to a substream containing the current offset to the end
		s->getSubStream(&ss, s->getOffset(), s->getLength() - s->getOffset());

		// Try to read attribute
		if (!a->read(&ss))
		{
			// Attribute read failed.
			// Clean up attribute
			delete a;

			// Return failure, indicating partial records.
			return false;
		}

		// a read successfully
		// End-of-attributes descriptor?
		if (a->type == 0xFFFFFFFF)
		{
			// Yes -- done reading attributes
			// Don't need to keep track of a, so get rid of it.
			delete a;

			// Return success.
			return true;
		}
		else
		{
			// a is a normal attribute
			// Add it to the attribute list
			m_list.append(a);

			// Make master stream reflect position of substream
			s->seek(s->getOffset() + ss.getOffset());

			// Keep reading.
		}
	}

	// End of attributes stream with no end-of-attributes descriptor
	// This isn't a good thing.
	WARN("FILE record missing terminating end-of-attributes descriptor.");
	return false;
}

NTFSAttribute * NTFSAttributes::getAttribute(NT32 type)
{
	NTFSAttribute * cur;
	for (cur = m_list.first(); cur; cur = m_list.next())
	{
		if (cur->type == type)
			return cur;
	}
	return NULL;
}

NTFSAttribute * NTFSAttributes::getAttributeById(NT32 type, NT16 id)
{
	NTFSAttribute * cur;
	for (cur = m_list.first(); cur; cur = m_list.next())
	{
		if (cur->type == type && cur->attributeId == id)
			return cur;
	}
	return NULL;
}

const QPtrList<NTFSAttribute> NTFSAttributes::getAttributesByType(NT32 type)
{
	QPtrList<NTFSAttribute> out;
	for (NTFSAttribute *cur = m_list.first(); cur; cur = m_list.next())
	{
		if (cur->type == type)
			out.append(cur);
	}
	return out;
}

QString NTFSAttributes::toString()
{
	if (m_list.first() == NULL)
	{
		return "Attributes: (none)";
	}

	QString out("Attributes:");
	NTFSAttribute * cur;

	for (cur = m_list.first(); cur; cur = m_list.next())
	{
		out += "\n\t" + cur->toString();
	}

	return out;
}

QString NTFSAttributes::getFileName()
{
	QString fileName;
	NT8 currentNameSpace;

	// Loop, looking for a file name
	NTFSAttribute * cur;
	for (cur = m_list.first(); cur; cur = m_list.next())
	{
		if (cur->type == 0x30)
		{
			// $FILE_NAME attribute
			NTFSAttributeDataFilename * fn;
			fn = (NTFSAttributeDataFilename *) cur->data;

			// See if it's more sensible than the previous name
			if (currentNameSpace == 0x02 || !fileName.length())
//			if (curName.length() > fileName.length() ||
//				(curName.length() == fileName.length() && curName.upper().compare(curName) != 0))
			{
				// curName is either longer or the same length and not completely uppercase
				// Set our return value to curName
				fileName = fn->fileName;
				currentNameSpace = fn->nameSpace;
			}
		}
	}

	// Done iterating
	return fileName;
}

bool NTFSAttributes::combineRunLists(const NTFSVolumeInfo * vi)
{
	QPtrList<NTFSAttribute> in, unhandled;
	NTFSAttribute * cur;

	in = m_list;
	in.setAutoDelete(false);
	unhandled.setAutoDelete(true);

	m_list.setAutoDelete(false);
	m_list.clear();
	m_list.setAutoDelete(true);

	// Copy our base attributes -- attributes with starting VCN of zero
	for (cur = in.first(); cur; cur = in.next())
	{
		if (!cur->nonResident || cur->vcnStart == 0)
		{
			// We have a base attribute
			m_list.append(cur);
		}
		else
		{
			// We don't have base attribute
			// Add it to the "unhandled" list
			unhandled.append(cur);
		}
	}

	// Given a list of unhandled attributes, tack their runlist on to the existing runlist

	// Algorithm: for each base attribute, iterate unhandled attributes looking for matching
	// type. Given matching type, find one with a starting VCN matching our ending VCN.
	// Upon match, append runlist, remove matching attribute, re-iterate. Upon no matching end
	// VCN but matching attribute, a warning is emitted and a sparse record is appended. Upon
	// no matching records, next base attribute is selected.
	//
	// Efficient? Not really. Works? Yes.
	NTFSAttribute * match;
	bool foundMatchingType;
	bool restartOuter = false;
	for (cur = m_list.first(); cur; cur = (restartOuter ? m_list.first() : m_list.next()))
	{
		// By default, do not restart outer list
		restartOuter = false;

		// Do we need more data?
		NT64 clustersNeeded = cur->sizeAlloc / vi->bytesPerCluster;
		if (!cur->nonResident || clustersNeeded == (cur->vcnStop + 1)) continue;

		// We need more data
		// Check the unhandled array

		// Look for a record that's part of the current record
		foundMatchingType = false;
		NT64 matchingVcnStart = (NT64)-1;
		for (match = unhandled.first(); match; match = unhandled.next())
		{
			// Is match our same type?
			if (match->type != cur->type) continue;

			// match matches our type
			// Flag as such
			foundMatchingType = true;

			// Record the minimum starting VCN
			matchingVcnStart = MIN(matchingVcnStart, match->vcnStart);

			// Does it match our current ending cluster?
			if (match->vcnStart != cur->vcnStop + 1) continue;

			// match begins at our current ending cluster
			// Append runlist.
			cur->runList.append(&match->runList);

			// Update stopping VCN -- increase by the length of match (cluster start/stop plus one)
			cur->vcnStop += (match->vcnStop - match->vcnStart) + 1;

			// Remove match
			unhandled.remove(match);
		}

		// We're done iterating. Did we find a matching type?
		if (foundMatchingType)
		{
			// Good. Start over.
			restartOuter = true;
		}
		else
		{
			// We need more data, but we didn't find it.
			// How many clusters are we missing?
			NT64 missingClusters = MIN(clustersNeeded, matchingVcnStart) - (cur->vcnStop + 1);

			// Emit a warning
			WARN(QString("Attribute type %1 requires %2 clusters that could not be located. Inserting sparse run.").arg(toHex(cur->type)).arg(toHex(missingClusters)));

			// Add sparse run
			cur->runList.addSparseRun(cur->vcnStop + 1, missingClusters);
			cur->vcnStop += missingClusters;

			restartOuter = true;
		}
	}

	// Success, probably
	// TODO: find out :-/

	return true;
}

bool NTFSAttributes::deReference(NTFSVolume * v)
{
	// Look for attribute lists
	QPtrList<NTFSAttribute> attrLists = getAttributesByType(0x20);
	NTFSAttribute * cur;
	for (cur = attrLists.first(); cur; cur = attrLists.next())
	{
		// cur points to an attribute list
		// Make sure it's de-referenced
		if (!cur->deReference(v))
		{
			// Failed to de-reference attribute list
			return false;
		}

		// Attribute list is de-referenced -- read its data
		NT32 cursor = 0;
		NTFSAttributeListEntry entry;
		NTFSStream substream;
		NTFSFile * file;
		NTFSAttributeDataAttribList * al;
		al = (NTFSAttributeDataAttribList *) cur->data;
		while (cursor < al->s.getLength())
		{
			// Get as substream for the entry to read
			al->s.getSubStream(&substream, cursor, al->s.getLength() - cursor);

			// Read the entry
			if (!entry.read(&substream))
			{
				// Failed to process attribute list entry
				STRACE("Failed to read attribute list entry.", &substream);
				return false;
			}

			// Advance the cursor
			cursor += substream.getOffset();

			// Do we need this attribute?
			//if (getAttributeById(entry.type, entry.attributeId) == NULL)
			if (1)
			{
				// Yes.
				// Read the file record pointed to by reference number
				file = v->getFileByReference(entry.extentReference);

				// Did we get a file?
				if (file != NULL)
				{
					// Got a file
					// Copy the attribute we want to ourselves
					NTFSAttribute * src = file->a.getAttributeById(entry.type, entry.attributeId);
					if (src)
					{
						// Found our attribute -- copy it
						NTFSAttribute * attr = new NTFSAttribute;
						attr->copy(src);
						m_list.append(attr);
					}
					else
					{
						SWARN(QString("De-referenced attribute list but did not locate attribute ID %1!")
							.arg(to16Hex(entry.attributeId)), &substream);
					}

					// Clean up
					delete file;
				}
				else
				{
					// Couldn't get a file for our attribute extent
					SWARN(QString("Unable to dereference file with reference number %1.")
						.arg(toHexRef(entry.extentReference)), &substream);
				} // get file?
			} // need attribute?
		} // attribute list entry loop

		// Remove $ATTRIBUTE_LIST from list of attributes
		m_list.remove(cur);

	} // attribute list loop

	// If we got here, we're either de-referenced or have no ATTRIBUTE_LIST
	// Either way, it's successful
	return combineRunLists(v->getVolInfo());
}
