// ntfsfile.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include "ntfsfile.h"
#include "ntfsutil.h"
#include "ntfsrunlist.h"

NTFSFile::NTFSFile()
{
	recordNum = 0;
	magic = 0;
	updateSequenceOffset = 0;
	updateSequenceSize = 0;
	logSequence = 0;
	recordSequence = 0;
	hardLinkCount = 0;
	attributeOffset = 0;
	flags = 0;
	recordSize = 0;
	allocatedSize = 0;
	baseFileRecord = 0;
	nextAttributeId = 0;

	m_s = NULL;
}

NTFSFile::~NTFSFile()
{
	if (m_s) delete m_s;
}

bool NTFSFile::read(NTFSStream * s)
{
	// Zero attributes
	a.zero();

	// Fixed-size header
	struct
	{
		NT32 magic;
		NT16 updateSequenceOffset;
		NT16 updateSequenceSize;
		NT64 logSequence;
		NT16 recordSequence;
		NT16 hardLinkCount;
		NT16 attributeOffset;
		NT16 flags;
		NT32 recordSize;
		NT32 allocatedSize;
	} head;
	// FIXME: under what situations will compilers add padding?

	// Read header
	s->readArray((char *)&head, sizeof(head));

	// Check for validity
	if (head.magic != 0x454c4946)
	{
		// No dice -- not a FILE record
		return false;
	}
	if (head.recordSize < 0x2A || head.recordSize > 0x1000)
	{
		// Shorter than minimum or longer than reasonable.
		// Likely not a FILE.
		//TRACE(QString("Invalid FILE record: recordSize = %1").arg(toHex(head.recordSize)));
		return false;
	}
	if (head.allocatedSize < 0x200 || head.allocatedSize > 0x1000)
	{
		// Again, shorter than minimum or longer than reasonable.
		// Probably not a FILE.
		STRACE(QString("Invalid FILE record: allocatedSize = %1").arg(toHex(head.allocatedSize)), s);
		return false;
	}
	if (head.allocatedSize < head.recordSize)
	{
		// Record is larger than allocated -- this doesn't occur in nature.
		// Probably not a FILE.
		STRACE(QString("Invalid FILE record: recordSize = %1, allocatedSize = %2")
			.arg(toHex(head.recordSize)).arg(toHex(head.allocatedSize)), s);
		return false;
	}
	if (head.allocatedSize != s->getLength())
	{
		// Allocated size is not the length of the stream
		// Either the value out of the boot sector is wrong, or this is.
		// We'll guess it's this one :-)
		STRACE(QString("Allocated size (%1) differs from size of stream (%2)")
			.arg(toHex(head.allocatedSize)).arg(toHex(s->getLength())), s);
	}

	// Things look good so far
	// Copy header data
	magic = head.magic;
	updateSequenceOffset = head.updateSequenceOffset;
	updateSequenceSize = head.updateSequenceSize;
	logSequence = head.logSequence;
	recordSequence = head.recordSequence;
	hardLinkCount = head.hardLinkCount;
	attributeOffset = head.attributeOffset;
	flags = head.flags;
	recordSize = head.recordSize;
	allocatedSize = head.allocatedSize;

	// Get our own FILE stream, corresponding to the record size, and use it
	m_s = new NTFSStream;
	s->getSubStream(m_s, 0, recordSize);

	// Save our record number
	recordNum = s->getVolumeRecord();

	// Process update sequence

	// First, see if we need to process update sequence
	NT32 bytesPerSector = m_s->getVolumeBytesPerSector();
	if (m_s->getLength() < bytesPerSector - 2)
	{
		// We don't intersect anything affected by the update sequence, so forget about it
	}
	else
	{
		// We hit a sector bound -- thus, we need to apply the update sequence

		// Read update sequence
		// Number and array storage
		NT16 updateSequenceNumber;
		NT16 * updateSequenceArray = new NT16[updateSequenceSize];

		// Seek to offset
		m_s->seek(updateSequenceOffset);

		// Read USN
		updateSequenceNumber = m_s->read16();

		// Read update sequence array
		if (!m_s->readArray((char*)updateSequenceArray, updateSequenceSize * 2))
		{
			SERROR("Failed to read update sequence.", m_s);
			delete[] updateSequenceArray;
			return false;
		}

		// Patch each sector bound
		NT32 offset;
		for (NT32 sector = 0; sector < (recordSize / bytesPerSector); sector ++)
		{
			// Offset that we're interested in
			offset = (sector + 1) * bytesPerSector - 2;

			// Point to that offset
			m_s->seek(offset);

			// Check that offset for the USN
			if (m_s->read16() != updateSequenceNumber)
			{
				// Update sequence is broken

				// Warn user
				SWARN("Update sequence is broken. Attempting to continue anyway.", m_s);

				// Try to process everything else anyway
				break;
			}

			// Patch our offset with the data from the updateSequenceArray
			m_s->patch(offset, (char *)&updateSequenceArray[sector], 2);
		}

		// Free updateSequenceArray
		delete[] updateSequenceArray;
	}

	// Update sequence has been applied, if necessary

	// More fields
	m_s->seek(0x20);
	baseFileRecord = m_s->read64();
	nextAttributeId = m_s->read16();

	// Read attributes
	m_s->seek(attributeOffset);
	a.read(m_s);

	return true;
}

bool NTFSFile::isValid()
{
	// Check magic descriptor
	if (magic != 0x454c4946)
	{
		// Not a FILE
		return false;
	}

	if (recordSize < 0x2A || recordSize > 0x1000)
	{
		// Record size shorter than minimum or unreasonably large
		return false;
	}

	if (allocatedSize < 0x200 || allocatedSize > 0x1000 || allocatedSize < recordSize)
	{
		// Record size shorter than minimum, unreasonably large, or nonsensical
		return false;
	}

	if (updateSequenceOffset > recordSize || attributeOffset > recordSize)
	{
		// Something doesn't make sense.
		return false;
	}

	return true;
}

QString NTFSFile::toString()
{
	if (!isValid())
		return "File {\n\t[invalid]\n}";

	QString out;

	// Standard things
	out = QString("File {\n\tRecord: %1\n\tRecord size: %2 / %3\n\tSequence number: %4")
		.arg(toHex(m_s->getVolumeRecord()))
		.arg(recordSize)
		.arg(allocatedSize)
		.arg(recordSequence);

	// Use extra nifty conditional output for non-normal things
	if (hardLinkCount != 1 && hardLinkCount != 2)
		out += QString("\n\tHard links: %1").arg(hardLinkCount);
	if (!(flags & 0x0001))
		out += "\n\tNot in use (i.e. deleted)";
	if (flags & 0x0002)
		out += "\n\tDirectory entry";
	if (baseFileRecord != 0)
		out += QString("\n\t$ATTRIBUTE_LIST extent\n\tBase file record: %1")
			.arg(toHexRef(baseFileRecord));

	// Write attributes
	out += "\n\t" + a.toString();

	// Close "File {"
	out += "\n}";

	return out;
}

bool NTFSFile::identify(NTFSStream * s)
{
	NT32 pos = s->getOffset();
	bool success = s->read32() == 0x454c4946;
	s->seek(pos);
	return success;
}
