// ntfsindex.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include "ntfsindex.h"
#include "ntfsutil.h"

NTFSIndex::NTFSIndex()
{
	indexEntries.setAutoDelete(true);
}

NTFSIndex::~NTFSIndex()
{
}

QString NTFSIndex::toString() const
{
	if (indexEntries.isEmpty()) return "Directory {\n\t(empty)\n}";

	QString out("Directory {\n");

	const NTFSIndexEntry * cur;
	QPtrListIterator<NTFSIndexEntry> it(indexEntries);

	while ((cur = it.current()) != NULL)
	{
		if (cur->data->nameSpace != 0x02)
			out.append(QString("\t%1\n").arg(cur->toString()));
		++it;
	}

	out.append("}");
	return out;
}

bool NTFSIndex::read(NTFSFile * f, NTFSVolume * v)
{
	// Sanity checks
	if (f == NULL)
	{
		BUG("Passed NULL file!");
		return false;
	}
	if (v == NULL)
	{
		BUG("Passed NULL volume!");
		return false;
	}

	// Clean up
	indexEntries.clear();

	NTFSAttribute * root, * alloc;

	// De-reference any $ATTRIBUTE_LISTs we might have
	if (!f->a.deReference(v))
	{
		ERROR("Failed to de-reference $ATTRIBUTE_LIST.");
		return false;
	}

	// Make sure f is a directory, by looking for $INDEX_ROOT
	root = f->a.getAttribute(0x90);
	if (root == NULL)
	{
		// Not a directory
		TRACE("File is not a directory: lacks $INDEX_ROOT attribute.");
		return false;
	}

	// De-reference $INDEX_ROOT (which it should be, already)
	if (!root->deReference(v))
	{
		// :-(
		ERROR("Unable to de-reference $INDEX_ROOT. (?)");
	}

	// Get root's data stream
	NTFSStream sRoot;
	root->getData(v, &sRoot);
	sRoot.seek(0);

	// Read index root stuff
	indexType = sRoot.read32();
	collationRule = sRoot.read32();
	recordSize = sRoot.read32();
	//TRACE(QString("Index record size = %1").arg(toHex(recordSize)));

	// Skip to index header stuff
	sRoot.seek(0x10);
	indexOffset = sRoot.read32();
	indexLength = sRoot.read32();
	allocatedLength = sRoot.read32();
	flags = sRoot.read8();

	// Determine index type
	if (flags & 0x01)
	{
		// Data is in a series of blocks inside $INDEX_ALLOCATION
		alloc = f->a.getAttribute(0xA0);
		if (alloc == NULL)
		{
			// Suck!
			ERROR("Directory is missing required $INDEX_ALLOCATION attribute.");
			return false;
		}

		// Get the index data
		NTFSStream sIndex;
		if (!alloc->getData(v, &sIndex))
		{
			ERROR("Failed to de-reference $INDEX_ALLOCATION attribute!");
			return false;
		}

		// sIndex contains a series of INDX records
		// Iterate through the records, using readBlock() on each of them
		NTFSStream ss;

		for (NT32 pos = 0; pos + recordSize <= sIndex.getLength(); pos += recordSize)
		{
			sIndex.getSubStream(&ss, pos, recordSize);
			if (!readBlock(&ss, v->getVolInfo()->bytesPerSector))
			{
				// :-(
				ERROR("Failed to read index block.");
				return false;
			}
		}

		// Everything seems okay.
		return true;
	}
	else
	{
		// We fit into the $INDEX_ROOT
		// Index entries begin at indexOffset and continue for indexLength
		// Note: indexOffset is relative to the beginning of the index header, which starts at 0x10

		// Seek to offset (shifted by 0x10)
		sRoot.seek(0x10 + indexOffset);

		// Read index entries
		NTFSIndexEntry * e;
		while (sRoot.getOffset() < sRoot.getLength() - 1)
		{
			e = new NTFSIndexEntry;
			if (e->read(&sRoot))
			{
				// Index entry read properly

				// Is it an index entry, or an end-of-index marker?
				if (e->data)
				{
					// Has data
					// Save it
					indexEntries.append(e);
				}
				else
				{
					// Read succeeded but produced no data -- must be end-of index
					// Exit read loop
					break;
				}
			}
			else
			{
				// :-(
				delete e;
				SERROR("Index entry read failed.", &sRoot);
				return false;
			}
		}
	}

	// We're done reading

	return true;
}

bool NTFSIndex::readBlock(NTFSStream * s, NT16 bytesPerSector)
{
	s->seek(0);
	if (s->read32() != 0x58444E49)
	{
		ERROR("Index block doesn't start with \"INDX\".");
		return false;
	}

	// Read update sequence info
	NT16 updateSequenceOffset = s->read16();
	NT16 updateSequenceSize = s->read16();

	// Apply update sequence
	if (updateSequenceSize)
	{
		// Read update sequence array
		NT16 * updateSequenceArray = new NT16[updateSequenceSize];

		// Seek to offset
		s->seek(updateSequenceOffset);

		// Read USN
		NT16 updateSequenceNumber = s->read16();

		// Read update sequence array
		if (!s->readArray((char*)updateSequenceArray, updateSequenceSize * 2))
		{
			SERROR("Failed to read update sequence.", s);
			delete[] updateSequenceArray;
			return false;
		}

		// Patch each sector bound
		NT32 offset;
		for (NT32 sector = 0; sector < (s->getLength() / bytesPerSector); sector ++)
		{
			// Offset that we're interested in
			offset = (sector + 1) * bytesPerSector - 2;

			// Point to that offset
			s->seek(offset);

			// Check that offset for the USN
			if (s->read16() != updateSequenceNumber)
			{
				// Update sequence is broken

				// Warn user
				SWARN("Update sequence is broken. Attempting to continue anyway.", s);

				// Try to process everything else anyway
				break;
			}

			// Patch our offset with the data from the updateSequenceArray
			s->patch(offset, (char *)&updateSequenceArray[sector], 2);
		}

		// Free updateSequenceArray
		delete[] updateSequenceArray;
	}

	// Update sequence is applied

	// Read info about our entries
	// Strangely, this information is relative to 0x18 -- compensate
	s->seek(0x18);
	NT32 entriesOffset = s->read32() + 0x18;
	NT32 entriesSize = s->read32() + 0x18;

	// Read entries
	NTFSIndexEntry * e;
	for (NT32 pos = entriesOffset; pos < entriesSize; pos = s->getOffset())
	{
		s->seek(pos);

		e = new NTFSIndexEntry;
		if (e->read(s))
		{
			// Read succeeded
			// Are we an end-of-index-entries entry?
			if (e->data)
			{
				// We have data
				indexEntries.append(e);
			}
			else
			{
				// No data; end-of-index entry
				delete e;
				break;
			}
		}
		else
		{
			// Read failed
			SERROR("Index entry read failed.", s);
			delete e;
			return false;
		}
	}

	return true;
}
