// ntfsrecovery.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include "ntfsrecovery.h"
#include "ntfsutil.h"

#include <qvaluelist.h>
#include <qdir.h>

#include <sys/types.h>
#include <utime.h>
#include <errno.h>

NTFSRecovery::NTFSRecovery()
{
	volumeFile = NULL;
	targetDir = ".";
}

NTFSRecovery::~NTFSRecovery()
{
	close();
}

bool NTFSRecovery::open(QString volume)
{
	// Make sure we're closed
	close();

	// Open volume
	volumeFile = fopen(volume.latin1(), "r");

	// Did it work?
	if (!volumeFile)
	{
		// Failed to open
		ERROR(QString("Unable to open volume file. Reason: %1").arg(strerror(errno)));
		return false;
	}

	// Attach NTFSVolume to volume file
	if (!v.attach(fileno(volumeFile)))
	{
		// Failure
		ERROR("Failed to attach to volume.");
		return false;
	}

	// We've attached
	// Save file name
	volumeFileName = volume;

	// Success
	return true;
}

void NTFSRecovery::close()
{
	if (volumeFile)
	{
		v.detach();
		fclose(volumeFile);
		volumeFile = NULL;
		volumeFileName = QString::null;
	}
}

bool NTFSRecovery::setRecoveryArea(QString target)
{
	// Make sure target directory exists
	QDir d;
	d.mkdir(target, true);

	// Set target directory
	targetDir = target;

	// chdir to it
	return d.cd(target, true);
}

bool NTFSRecovery::recoverFile(NT64 record, QString path)
{
if (path.length() == 0) path = targetDir;

	// Get FILE at record
	NTFSFile * f;
	f = v.getFile(record);

	// Did it work?
	if (f == NULL)
	{
		// No :-(
		ERROR(QString("Failed to read FILE at record %1.").arg(toHex(record)));
		return false;
	}

	// f points to a valid file
	// See if it's a directory
	if (f->isDirectory())
	{
		// Yep, it's a directory
		// Call recoverDirectory instead
		delete f;
		return recoverDirectory(record, path);
	}
	else
	{
		// It's a file
		// Find its name
		QString fileName = f->a.getFileName();

		// Make up a name if we didn't get one
		if (fileName.length() == 0)
			fileName = toHex(record);

		// De-reference everything
		f->a.deReference(&v);

		// Open output file
		QString outputFileName = QString("%1/%2").arg(path).arg(fileName);
		FILE * o = fopen(outputFileName.latin1(), "w");

		// Get our data attribute
		NTFSAttribute * data = f->a.getAttribute(0x80);

		// Did we get data?
		bool success;
		if (data)
		{
			// Got data attribute
			success = data->writeContentsToFile(&v, fileno(o));
		}
		else
		{
			// No data :-(
			success = false;
			ERROR(QString("\"%1\" (record %2) has no data attribute.")
				.arg(fileName).arg(toHex(record)));
		}

		// Close output file
		fclose(o);

		// Copy our file times
		NTFSAttribute * stdinfo = f->a.getAttribute(0x10);
		if (stdinfo)
		{
			// Get standard information
			NTFSAttributeDataStdInfo * info;
			info = (NTFSAttributeDataStdInfo *) stdinfo->data;

			// Fill a utimbuf with data
			utimbuf b;
			b.actime = time64ToUnix(info->aTime);
			b.modtime = time64ToUnix(info->mTime);

			// Set the time, ignoring errors, since it's not really necessary
			utime(outputFileName.latin1(), &b);
		}

		// Clean up
		delete f;

		// Return success
		return success;
	}
}

bool NTFSRecovery::recoverDirectory(NT64 record, QString path)
{
	if (path.length() == 0) path = targetDir;

	// Get FILE at record
	NTFSFile * f;
	f = v.getFile(record);

	// Did it work?
	if (f == NULL)
	{
		// No :-(
		ERROR(QString("Failed to read FILE at record %1.").arg(toHex(record)));
		return false;
	}

	// f points to a valid file
	QString fileName = f->a.getFileName();

	TRACE(QString("Recovering \"%1\" into %2.").arg(fileName).arg(path));

	// See if it's a directory
	if (!f->isDirectory())
	{
		// Not a directory
		ERROR(QString("File \"%1\" (record %2) is not a directory.").arg(fileName).arg(toHex(record)));
		delete f;
		return false;
	}
	else
	{
		// It's a directory

		// Make up a name if we didn't get one
		if (fileName.length() == 0)
			fileName = toHex(record);

		// De-reference everything
		f->a.deReference(&v);

		// Attach an index
		NTFSIndex i;
		if (!i.read(f, &v))
		{
			ERROR(QString("File \"%1\" (record %2) had no index.").arg(fileName).arg(toHex(record)));
			delete f;
			return false;
		}

		// Index got attached, so proceed.

		// Create output directory
		QString outputFileName = QString("%1/%2").arg(path).arg(fileName);
		QDir d;
		if (!d.mkdir(outputFileName, true))
		{
			ERROR(QString("Failed to create output directory \"%1\".").arg(outputFileName));
			return false;
		}

		// Get a list of files to recover
		QValueList<NT64> references;
		{
			const NTFSIndexEntry * cur;
			QPtrListIterator<NTFSIndexEntry> it(i.indexEntries);

			while ((cur = it.current()) != NULL)
			{
				if (!references.contains(cur->fileRef))
					references.append(cur->fileRef);
				++it;
			}
		}

		// references contains a bunch of file references
		// Recover all of them, specifying our output directory as their path
		QValueList<NT64>::iterator it;
		NT64 record;
		for (it = references.begin(); it != references.end(); ++it)
		{
			record = v.getVolInfo()->getRecordForReference(*it);
			if (!recoverFile(record, outputFileName))
			{
				// :-(
				ERROR(QString("Recovery of file record %1 (within \"%2\") failed.")
					.arg(toHex(*it)).arg(outputFileName));
			}
		}

		// Copy our file times
		NTFSAttribute * stdinfo = f->a.getAttribute(0x10);
		if (stdinfo)
		{
			// Get standard information
			NTFSAttributeDataStdInfo * info;
			info = (NTFSAttributeDataStdInfo *) stdinfo->data;

			// Fill a utimbuf with data
			utimbuf b;
			b.actime = time64ToUnix(info->aTime);
			b.modtime = time64ToUnix(info->mTime);

			// Set the time, ignoring errors, since it's not really necessary
			utime(outputFileName.latin1(), &b);
		}

		// Clean up
		delete f;

		// Return success
		return true;
	}
}
