// ntfsstream.h
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include "ntfsvolume.h"
#include "ntfsstream.h"
#include "ntfsutil.h"

// write()
#include <unistd.h>

NTFSStream::NTFSStream()
{
	// Intialize members
	m_data = NULL;
	m_volume = NULL;
	m_dataLen = m_position = m_dataLen = 0;
}

NTFSStream::~NTFSStream()
{
	// Free resources, if any
	detach();
}

void NTFSStream::attach(char * data, NT32 length, NT64 position, NTFSVolume * parent)
{
	// Detach, if we're not detached already
	detach();

	// Save pointer to data
	m_data = data;

	// Save length and position
	m_dataLen = length;
	m_position = position;

	// Set parent volume
	m_volume = parent;

	// Tell parent volume, if any, that we're using its data
	if (m_volume)
		m_volume->addChildStream(this);

	// Seek to origin
	seek(0);
}

void NTFSStream::attachCopy(const char * data, NT32 length, NT64 position)
{
	// Create a copy
	char * mydata = new char[length];
	memcpy(mydata, data, length);

	// Let attach() do the rest
	attach(mydata, length, position, NULL);
}

void NTFSStream::detach()
{
	// Are we using a volume's data?
	if (m_volume)
	{
		// Tell it we're no longer using its data
		m_volume->removeChildStream(this);

		// Forget about our volume
		m_volume = NULL;
	}
	else if (m_data)
	{
		// We own our own data
		// delete[] it.
		delete[] m_data;
	}

	// Zero our data pointer
	m_data = NULL;

	// Zero length and cursor
	m_dataLen = m_cursor = 0;
}

void NTFSStream::unbind()
{
	// Volume is cleaning the cache, we're doing a write, or something.
	// See if we're bound
	if (m_volume)
	{
		// Yeah, volume owns data.
		// Since we want to keep our data, assign ourselves to a copy of it.

		// Save cursor
		NT32 cursor = m_cursor;

		// Make the copy
		attachCopy(m_data, m_dataLen, m_position);

		// Restore cursor
		m_cursor = cursor;
	}
}

bool NTFSStream::getSubStream(NTFSStream *target, NT32 offset, NT32 length) const
{
	// Validate inputs
	if (target == NULL) return false;
	if (offset + length > m_dataLen)
	{
		// Don't have that much data!
		return false;
	}

	if (m_volume)
	{
		// Attach to a subsection of volume's data
		target->attach(&m_data[offset], length, m_position + offset, m_volume);
	}
	else
	{
		// Attach to a copy of a subsection of our data
		target->attachCopy(&m_data[offset], length, m_position + offset);
	}

	// Set substream's volume info to our volume info
	target->m_volInfo = m_volInfo;

	return true;
}

bool NTFSStream::getSubStreamCopy(NTFSStream *target, NT32 offset, NT32 length) const
{
	// Validate inputs
	if (target == NULL) return false;
	if (offset + length > m_dataLen)
	{
		// Don't have that much data!
		return false;
	}

	// Use attachCopy() on a subsection of our data
	target->attachCopy(&m_data[offset], length, m_position + offset);

	// Set target's volume info to our volume info
	target->m_volInfo = m_volInfo;

	return true;
}

// Read an NT64 from the stream
NT64 NTFSStream::read64()
{
	if (m_cursor + 8 > m_dataLen)
	{
		SBUG("Read beyond end of stream.", this);
		return 0;
	}

	union {
		NT8 data[8];
		NT64 u64;
	} conv;

	conv.data[0] = m_data[m_cursor++];
	conv.data[1] = m_data[m_cursor++];
	conv.data[2] = m_data[m_cursor++];
	conv.data[3] = m_data[m_cursor++];
	conv.data[4] = m_data[m_cursor++];
	conv.data[5] = m_data[m_cursor++];
	conv.data[6] = m_data[m_cursor++];
	conv.data[7] = m_data[m_cursor++];

	return conv.u64;
}

// Read an NT32 from the stream
NT32 NTFSStream::read32()
{
	if (m_cursor + 4 > m_dataLen)
	{
		SBUG("Read beyond end of stream.", this);
		return 0;
	}

	union {
		NT8 data[4];
		NT32 u32;
	} conv;

	conv.data[0] = m_data[m_cursor++];
	conv.data[1] = m_data[m_cursor++];
	conv.data[2] = m_data[m_cursor++];
	conv.data[3] = m_data[m_cursor++];

	return conv.u32;
}

// Read an NT16 from the stream
NT16 NTFSStream::read16()
{
	if (m_cursor + 2 > m_dataLen)
	{
		SBUG("Read beyond end of stream.", this);
		return 0;
	}

	union {
		NT8 data[2];
		NT16 u16;
	} conv;

	conv.data[0] = m_data[m_cursor++];
	conv.data[1] = m_data[m_cursor++];

	return conv.u16;
}

// Read an NT8 from the stream
NT8 NTFSStream::read8()
{
	if (m_cursor + 1 > m_dataLen)
	{
		SBUG("Read beyond end of stream.", this);
		return 0;
	}

	return m_data[m_cursor++];
}

// Read a Unicode string from the stream
QString NTFSStream::readString(NT16 length)
{
	// Return an empty string if length is empty
	if (length == 0) return QString::null;

	// Allocate an NT16 array
	NT16 * data = new NT16[length];

	// Read into unicode character array
	readArray((char*)data, length * 2);

	// Copy into QString
	QString out((QChar*)data, length);

	// Free NT16 array
	delete[] data;

	// Send string to caller
	return out;
}

// Read an ASCII string from the stream
QString NTFSStream::readAscii(NT16 length)
{
	// Return an empty string if length is empty
	if (length == 0) return QString::null;

	// Allocate an NT16 array
	QChar * data = new QChar[length];

	// Read ASCII into unicode character array
	for (int i = 0; i < length; i ++)
		data[i] = read8();

	// Copy into QString
	QString out(data, length);

	// Free NT16 array
	delete[] data;

	// Send string to caller
	return out;
}

// Reads a GUID and formats MS registry style
QString NTFSStream::readGuidAsString()
{
	NT16 guid[8];
	for (int i = 0; i < 8; i ++)
		guid[i] = read16();

	QString out;
	out.sprintf("{%x%x-%x-%x-%x-%x%x%x}",
		guid[0], guid[1], guid[2], guid[3], guid[4], guid[5], guid[6], guid[7]);

	return out;
}

// Fill dest[0] to dest[length] with data from cursor
bool NTFSStream::readArray(char * dest, NT32 length)
{
	if (m_data == NULL)
	{
		BUG("Stream is closed!");
		return false;
	}

	// Check that dest is set
	if (dest == NULL)
	{
		BUG("ERRORCaller specified NULL destination pointer.");
		return false;
	}

	// Get cursor
	NT32 pos = getOffset();

	// Check that what we want to do is valid
	if (pos + length > getLength())
	{
		// Too long
		SBUG(QString("Requested size (%1) greater than remaining stream length (%2)")
			.arg(length).arg(length - pos), this);
		return false;
	}

	// Everything checks out -- memcpy!
	memcpy(dest, m_data + pos, length);

	// Advance cursor
	seek(pos + length);

	// Success
	return true;
}

bool NTFSStream::readByteArray(QByteArray &arr, NT32 length)
{
	if (arr.size() < length) arr.resize(length);
	return readArray(arr.data(), length);
}

// Update a chunk of data
void NTFSStream::patch(NT32 offset, const char * data, NT32 length)
{
	// We're writing; volume's data is read-only, so make sure we have our own data
	unbind();

	// Force length to be within the bounds of the stream
	length = (offset + length > m_dataLen ? m_dataLen - offset : length);

	// Write data to offset... simple enough
	memcpy(m_data + offset, data, length);
}

bool NTFSStream::writeToFile(int fd, NT32 length) const
{
	if (m_data == NULL)
	{
		BUG("Stream is closed!");
		return false;
	}

	// Check that dest is set
	if (fd == 0)
	{
		BUG("Caller specified invalid file descriptor.");
		return false;
	}

	// Check that what we want to do is valid
	if (length > getLength())
	{
		// Too long
		SBUG(QString("Requested size (%1) greater than stream length (%2)")
			.arg(length).arg(getLength()), this);
		return false;
	}

	// Everything checks out -- write() directly
//	TRACE(QString("Writing %1 bytes.").arg(toHex(length)));
	NT32 writtenBytes = write(fd, m_data, length);

	// Success?
	return (writtenBytes == length);
}
