// ntfsvolume.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

// For mmap()ing
#include <stdio.h>
#include <sys/mman.h>
#include <errno.h>

// For write()
#include <unistd.h>

#include "ntfsvolume.h"
#include "ntfsutil.h"
#include "ntfsfile.h"

NTFSVolume::NTFSVolume()
{
	m_device = 0;
	cachedBlocks.setAutoDelete(true);
	useMmap = true;
}

NTFSVolume::~NTFSVolume()
{
	// Cleanly detach from m_device, if any.
	detach();
}

bool NTFSVolume::attach(int fd)
{
	// Assign local device pointer to the device
	m_device = fd;

	// Success
	return true;
}

void NTFSVolume::detach()
{
	// Unmap any mapped clusters
	manageCache(0);
}

char * NTFSVolume::cache(NT64 pos, NT32 length)
{
	MAPPEDCLUSTER * cur;

	// See our requested position is already cached
//	TRACE(QString("Searching for %1 bytes at %2").arg(toHex(length)).arg(toHex(pos)));
	for (cur = cachedBlocks.first(); cur; cur = cachedBlocks.next())
	{
		NT64 curBegin = cur->pos;
		NT64 curEnd = curBegin + cur->length;
//		TRACE(QString("Cache entry: %1 bytes at %2").arg(cur->length).arg(toHex(cur->pos)));
		if (pos >= curBegin && pos + length <= curEnd)
		{
			// Hit!
//			TRACE(QString("Cache hit at base address %1. Returned range ends at %2 of %3.")
//				.arg(toHex((NT32)cur->base)).arg(toHex(pos + length - curBegin)).arg(toHex(cur->length)));
			return (cur->base + (pos - curBegin));
		}
	}

	// We want to map in chunks of 4 MB, or 2^22 bytes
	NT16 chunkShift = 22;

	// Break into the chunk's base address and a byte offset
	NT64 chunkPos = (pos >> chunkShift) << chunkShift;
	NT32 byteOffset = (NT32)(pos - chunkPos);

	// Figure out the number of chunks needed to satisfy length
	NT32 chunks = ((byteOffset + length) >> chunkShift) + 1;

	// ...and the length of those chunks
	NT32 chunkLen = chunks << chunkShift;

	if (chunks == 0 || chunkPos > pos)
	{
		BUG("AIEEE!!!!");
		return NULL;
	}

	// Make a new mapped entry
	cur = new MAPPEDCLUSTER;

	// Set known fields
	cur->length = chunkLen;
	cur->pos = chunkPos;

	// Use the favored map technique
	cur->base = (char *)mapData(chunkPos, chunkLen);

	// Check if it's valid
	if (cur->base != NULL)
	{
		// Valid
		// Add to list of mapped clusters
		cachedBlocks.append(cur);

		// Return requested offset into the cluster
		return (cur->base + byteOffset);
	}
	else
	{
		// Not valid
		// Dump error
		ERROR(QString("Unable to cache() %1 bytes at offset %2! Error: %3")
			.arg(chunkLen).arg(toHex(chunkPos)).arg(strerror(errno)));

		// Free mapped resource
		delete cur;

		return NULL;
	}
}

void NTFSVolume::manageCache(NT16 cacheEntries)
{
	MAPPEDCLUSTER * cur;
	NTFSStream * s;

	// Make sure we have twelve items or less :-)
	int cacheSize = cachedBlocks.count();
	while (cacheSize > cacheEntries)
	{
		// TODO: stream counting :-(

		// Get the oldest mapped block
		cur = cachedBlocks.getFirst();

		// Unbind any and all attached streams
		for (s = cur->children.first(); s; s = cur->children.next())
			s->unbind();

		// Unmap memory
		unmapData(cur->base, cur->length);

		// Remove the block from our list
		cachedBlocks.removeFirst();

		// Update our copy of the cache's size
		cacheSize --;
	}
}

void NTFSVolume::addChildStream(NTFSStream * s)
{
	// Find the mapped block that s corresponds to by checking its data pointer
	MAPPEDCLUSTER * cur;
	for (cur = cachedBlocks.first(); cur; cur = cachedBlocks.next())
	{
		if (cur->base <= s->m_data && s->m_data <= (cur->base + cur->length))
		{
			// Stream's data pointer (s->m_data) is between the base address and the end address
			// That is, this stream is referencing this block
			// Add it to the list of child streams
			cur->children.append(s);

			// Bail out, since it's not going to match any others
			return;
		}
	}

	// This is bad. We didn't find a mapped block that s corresponds to!
	BUG("Unable to locate corresponding block for passed stream!");
}

void NTFSVolume::removeChildStream(NTFSStream * s)
{
	// Find the stream in our currently mapped blocks
	MAPPEDCLUSTER * cur;
	for (cur = cachedBlocks.first(); cur; cur = cachedBlocks.next())
	{
		if (cur->children.remove(s))
		{
			// We removed our stream, yo!
			return;
		}
	}

	// If we got here, then we didn't find our stream anywhere.
	// Well, that's as close to de-referencing as we're going to get, isn't it? ;-)
}

bool NTFSVolume::streamData(NTFSStream *stream, NT64 pos, NT32 length)
{
	// Bail if no stream
	if (stream == NULL)
	{
		BUG("NULL stream passed.");
		return false;
	}
	// Bail if zero length
	if (length == 0)
	{
		BUG("Caller requested zero-length read.");
		return false;
	}

	// Get an address for our offset
	char * base = cache(pos, length);
	if (base == NULL)
	{
		ERROR(QString("Unable to cache() position %1 for length %2.").arg(toHex(pos)).arg(toHex(length)));
	}

	// Everything looks good -- attach stream to data
	stream->attach(base, length, pos, this);

	// Set stream's volume info pointer to our volume info
	stream->m_volInfo = &m_info;

	// Mop up our cache
	manageCache();

	// Success
	return true;
}

// Create a new stream, attach it to data at given pos with given length
//   Upon success, return stream
//   Upon failure, return NULL
NTFSStream * NTFSVolume::streamData(NT64 pos, NT32 length)
{
	NTFSStream * s = new NTFSStream();
	if (streamData(s, pos, length) == true)
	{
		// Success!
		return s;
	}
	else
	{
		// Failure :-(
		delete s;
		return NULL;
	}
}

bool NTFSVolume::analyzeBootSector(NT64 sector)
{
	// Assuming sector size of 512 bytes.
	// This is a safe assumption -- everything we need is in the first 512 bytes,
	// and an iterative scan will always locate the boot sector successfully, even
	// if it's location in sectors isn't accurate. (That's okay.)

	// Read 512 bytes from sector offset into stream object
	NTFSStream s;
	if (!streamData(&s, sector * 512, 512))
	{
		// Didn't get a stream (?!)
		ERROR(QString("Failed to read sector %1!").arg(toHex(sector)));
		return false;
	}

	// Check for "NTFS    " header
	s.seek(0x03);
	if (QString::compare(s.readAscii(8), "NTFS    ") != 0)
	{
		// Header isn't as expected
		return false;
	}

	// Read data
	s.seek(0x0B);
	NT16 bytesPerSector = s.read16();
	NT8 sectorsPerCluster = s.read8();
	s.seek(0x28);
	NT64 sectorCount = s.read64();
	NT64 offsetMft = s.read64();
	NT64 offsetMftMirr = s.read64();
	NT16S clusterPerFile = s.read16s();
	NT16S clusterPerIndex = s.read16s();
	NT64 serialNumber = s.read64();

	// Ensure everything makes sense
	// BPS should be 512 or 1024
	if (bytesPerSector != 512 && bytesPerSector != 1024)
	{
		ERROR(QString("%1 bytes per sector is invalid.").arg(bytesPerSector));
		return false;
	}

	// SPC should be 1, 2, 4, 8, 16, 32, 64, or maybe 128
	if (sectorsPerCluster != 1 && sectorsPerCluster != 2 && sectorsPerCluster != 4 &&
		sectorsPerCluster != 8 && sectorsPerCluster != 16 && sectorsPerCluster != 32 &&
		sectorsPerCluster != 64 && sectorsPerCluster != 128)
	{
		ERROR(QString("%1 bytes per sector is invalid.").arg(bytesPerSector));
		return false;
	}

	// Everything seems to look okay

	// Compute immediate
	NT32 bytesPerCluster = bytesPerSector * sectorsPerCluster;

	// Turn clusterPer* into bytesPer*
	NT32 bytesPerFile, bytesPerIndex;

	if (clusterPerFile < 0)
	{
		// Negative amount: 2 ^ -amt bytes
		bytesPerFile = 2 ^ (-clusterPerFile);
	}
	else
	{
		// Positive amount: bytes/cluster * amt bytes
		bytesPerFile = bytesPerCluster * clusterPerFile;
	}
	if (clusterPerIndex < 0)
	{
		// Negative amount: 2 ^ -amt bytes
		bytesPerIndex = 2 ^ (-clusterPerIndex);
	}
	else
	{
		// Positive amount: bytes/cluster * amt bytes
		bytesPerIndex = bytesPerCluster * clusterPerIndex;
	}

	// Fill in m_info with new fields
	m_info.bytesPerSector = bytesPerSector;
	m_info.sectorsPerCluster = sectorsPerCluster;
	m_info.sectorCount = sectorCount;
	m_info.offsetToMft = offsetMft;
	m_info.offsetToMftMirr = offsetMftMirr;
	m_info.bytesPerRecord = bytesPerFile;
	m_info.bytesPerIndex = bytesPerIndex;
	m_info.serialNumber = serialNumber;
	m_info.bytesPerCluster = bytesPerCluster;

	// Have m_info re-calculate internal fields
	m_info.recalculate();

	TRACE(QString("Boot sector identified at sector %1, reporting cluster size of %2 bytes.")
			.arg(toHex(sector))
			.arg(m_info.bytesPerCluster)
		);

	return true;
}

NTFSFile * NTFSVolume::getFile(NT64 record)
{
	// Read record into a local stream
	NTFSStream s;
	streamRecord(&s, record);

	// Have NTFSFile take a quick look
	if (NTFSFile::identify(&s))
	{
		// Seems promising
		// Create a new file object
		NTFSFile * f = new NTFSFile;

		// Read from stream
		s.seek(0);
		if (f->read(&s))
		{
			// Success
			return f;
		}

		// Failure
		// Clean up file object
		delete f;
	}

	// Something didn't end up working right
	// Die
	return NULL;
}

// Copies data from pos with length to file descriptor fd
bool NTFSVolume::copyData(NT64 pos, NT64 length, int fd)
{
	// Fixed 512 KB buffer
	char buffer[512 * 1024];

	// Seek to our position
	if (lseek64(m_device, pos, SEEK_SET) == (off_t)-1)
	{
		// Seek failed
		ERROR(QString("Could not seek to position %1.").arg(toHex(pos)));
		return false;
	}

	// Iteratively read into buffer and write to file descriptor
	NT64 lengthRemaining = length;
	NT32S curLength;
	while (lengthRemaining > 0)
	{
		// Length of current block
		curLength = (sizeof(buffer) > lengthRemaining ? (NT32S)lengthRemaining : sizeof(buffer));

		// Read data
		if (read(m_device, buffer, curLength) != curLength)
		{
			// Read failed
			ERROR(QString("Read at position %1 failed.").arg(toHex(pos + length - lengthRemaining)));
			return false;
		}

		// Write data
		if (write(fd, buffer, curLength) != curLength)
		{
			// Write failed
			ERROR(QString("Write of data from position %1 failed.")
				.arg(toHex(pos + length - lengthRemaining)));
			return false;
		}

		// Update length remaining
		lengthRemaining -= curLength;
	}

	// Success
	return true;
}

void * NTFSVolume::mapData(NT64 offset, NT32 length)
{
	if (useMmap)
	{
		// Memory-map
		void * addr = mmap(0, length, PROT_READ, MAP_SHARED, m_device, offset);

		TRACE(QString("mmap(0, %1, PROT_READ, MAP_SHARED, %2, %3) = %4")
			.arg(toHex(length)).arg(toHex((NT32)m_device))
			.arg(toHex((NT64)((off_t)offset))).arg(toHex((NT32)addr)));

		// Return the address, or NULL if failure
		return (addr == MAP_FAILED ? NULL : addr);
	}
	else
	{
		// Use read() syscall

		// Allocate memory
		char * buffer = new char[length];

		// Seek
		if (lseek64(m_device, offset, SEEK_SET) == (off_t)-1)
		{
			// Seek failed
			ERROR(QString("Could not seek to position %1.").arg(toHex(offset)));
			delete[] buffer;
			return NULL;
		}

		// Read into buffer
		if ((NT32)read(m_device, buffer, length) != length)
		{
			// Read failed
			ERROR(QString("Read at position %1 failed.").arg(toHex(offset)));
			delete[] buffer;
			return NULL;
		}

		return buffer;
	}
}

void NTFSVolume::unmapData(void * addr, NT32 length)
{
	if (useMmap)
	{
		// Unmap address
		munmap(addr, length);
	}
	else
	{
		// Free memory
		delete[] (char *)addr;
	}
}
